#!/usr/bin/python

''' Git Central Repo Monitor
    Check whether the contents of a target directory and its subdirectories,
    other than the contents of Git repositories, have changed since the
    previous run.

    See parse_args() or run with --help for usage.
'''

from __future__ import print_function

import csv
import errno
import grp
import optparse
import os
import pwd
import re
import stat
import subprocess
import sys
import tempfile
import urllib
from datetime import datetime

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
DEFAULT_RECDIR = '/usr/local/var/gcrm'

# Warning codes

W_OSERROR = 1
W_STARTEXCLUDE = 11
W_STOPEXCLUDE = 12
W_BADGROUP = 21
W_BADMODE = 22
W_NONBAREREPO = 31
W_BADBAREREPO = 32
W_UNSHAREDREPO = 33

W_MESSAGE = {
        W_OSERROR:      "OS error",
        W_STARTEXCLUDE: "started excluding directory contents",
        W_STOPEXCLUDE:  "stopped excluding directory contents",
        W_BADGROUP:     "bad group ownership",
        W_BADMODE:      "bad directory mode (should be group rws)",
        W_NONBAREREPO:  "non-bare Git repository",
        W_BADBAREREPO:  "bare Git repository should end with .git",
        W_UNSHAREDREPO: "unshared Git repository",
}

W_DETAIL = {
        W_OSERROR: """\
An error occurred during an OS call--probably stat or listdir. Inspect the
files based on the errno message.""",
        W_STARTEXCLUDE: """\
Directory contents which were previously included are now excluded. This just
shows the results of a change to the command used to run the report. It will
not appear on future runs.""",
        W_STOPEXCLUDE: """\
Directory contents which were previously excluded are now included. This just
shows the results of a change to the command used to run the report. It will
not appear on future runs.""",
        W_BADGROUP: """\
A file or directory is not owned by the proper group (specified on the command
line). Fix with (recursive) chown.""",
        W_BADMODE: """\
A directory is not group read-write-executable or does not have the SGID bit
set. Fix with chmod. You'll also need to make sure that files and
subdirectories have the correct group and permissions (including SGID on
subdirectories).""",
        W_NONBAREREPO: """\
A Git repository is not configured as bare (without a work tree). Fix with
"git config core.bare true" in the repo directory. For consistency, you may
also need to move the contents of a ".git" subdirectory to the repo directory
and remove the now-empty .git directory.""",
        W_BADBAREREPO: """\
A bare Git repository's name does not end with ".git", as is customary.
Renaming a Git repository generally needs to be negotiated with the owner and
anyone else who has a local repo with this one as a remote, since they will
need to use "git remote set-url" (see man git-remote) to point to the new
name.  However, if the repo just needs to be renamed from "something" to
"something.git", this change can be made without breaking local repos, since
Git will look for "something.git" after failing to find "something". You
should still talk to the owner, just in case there are any quirks. The rename
can be done with mv.""",
        W_UNSHAREDREPO: """\
A Git repository is not configured as shared; this means that Git will not
maintain group-write permissions on the files and directories that need them
for all members of the group to push to the repo. Fix with
"git config core.sharedrepository 1" in the repo directory. You should also
check the permissions of files and subdirectories: directories should be 2775,
and files should be either 444 (objects and packs, which are read-only) or
664.""",
}


# File data

def is_git(path):
    '''Return a tuple of bools (repo, bare, shared).

    repo iff path is a repo or a directory controlled by one.
    bare iff path is a bare repo.
    shared iff path is a shared repo.
    '''

    # Run git config in the directory and check the return code.
    # If it fails, then not a git repo. Otherwise, it is (for our purposes).
    # If it is a repo, the output will tell us whether it is bare/shared.
    try:
        # $GIT_DIR can confuse git config, so remove it from the environment
        # LBYL only because it's usually not there
        if 'GIT_DIR' in os.environ:
            del os.environ['GIT_DIR']

        # Sample output of command (run from a bare, shared repo):

        # core.bare true
        # core.sharedrepository 1

        # If a config option isn't defined, it isn't printed. These two default
        # to false, so if they don't print, they're false.
        p = subprocess.Popen(
                ['git', 'config', '--get-regex',
                    'core.(bare|sharedrepository)'],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=path
        )
        stdout, stderr = p.communicate()
        repo = not p.returncode
        bare = shared = False

        truevals = ('yes', '1', 'true', 'on')
        for line in stdout.splitlines():
            key, val = map(lambda s: s.strip().lower(), line.split(None, 1))
            if key == 'core.bare':
                bare = val in truevals
            elif key == 'core.sharedrepository':
                shared = val in ('group',) + truevals

        return repo, bare, shared
    except OSError:
        return False, False, False


class FileData(object):
    '''Holds file metadata.

    Important attribute names are in FIELDS.

    type -- one of 'file', 'dir', 'git'
    path -- filesystem path
    pruned -- whether a directory was pruned

    The rest refer to values returned from stat().
    '''

    # CSV headers
    FIELDS = ['type', 'path', 'mode', 'uid', 'gid', 'ctime', 'pruned']

    def __init__(self, path):
        '''Initializes self from stat(path).'''
        st = os.stat(path)

        self.path = path
        self.mode = stat.S_IMODE(st.st_mode)
        self.uid = st.st_uid
        self.gid = st.st_gid
        self.ctime = int(st.st_ctime)
        self.pruned = False
        if stat.S_ISDIR(st.st_mode):
            repo, bare, shared = is_git(path)
            if repo:
                self.type = 'git'
                self.bare = bare
                self.shared = shared
            else:
                self.type = 'dir'
        else:
            self.type = 'file'

    def __str__(self):
        '''Returns str of self to appear in notifications.'''
        if hasattr(self, '_str'):
            return self._str

        try:
            owner = pwd.getpwuid(self.uid).pw_name
        except KeyError:
            owner = self.uid
        try:
            group = grp.getgrgid(self.gid).gr_name
        except KeyError:
            group = self.gid
        time = datetime.fromtimestamp(self.ctime).strftime('%Y-%m-%d %H:%M:%S')

        self._str = ("%s (%s, %04o, %s:%s, changed %s)" %
                     self.path, self.type, self.mode, owner, group, time)
        return self._str

    def isdir(self):
        '''Returns true if self is for a directory, false otherwise.'''
        return self.type in ('dir', 'git')

    def get_warnings(self, gid=None):
        '''Returns a dict { warningcode: ['warningstr'] } for self.'''
        if hasattr(self, '_warnings'):
            return self._warnings

        self._warnings = {}

        if self.type == 'git':
            if not self.bare:
                self._warnings[W_NONBAREREPO] = [str(self)]
            elif not self.path.endswith('.git'):
                self._warnings[W_BADBAREREPO] = [str(self)]
            if not self.shared:
                self._warnings[W_UNSHAREDREPO] = [str(self)]

        if self.isdir():
            # Is mode at least 2070 == d---rws--- == sgid + group rwx?
            # bool(~i & j) == there is a bit set in j but not in i
            if ~stat.S_IMODE(self.mode) & (stat.S_ISGID | stat.S_IRWXG):
                self._warnings[W_BADMODE] = [str(self)]

        if gid and self.gid not in gid:
            self._warnings[W_BADGROUP] = [str(self)]

        return self._warnings

    @classmethod
    def fromfields(cls, fields):
        '''Returns a FileData object with attribute values from fields.

        Can be used with a csv.DictReader to reload a FileData from CSV.
        '''

        # bypass __init__ by calling __new__
        rv = object.__new__(cls)
        rv.path = fields['path']
        rv.type = fields['type']
        rv.pruned = (fields['pruned'] == 'True')
        for f in ('mode', 'uid', 'gid', 'ctime'):
            setattr(rv, f, int(fields[f]))
        return rv

    def tofile(self, csvout):
        '''Writes self to csv.DictWriter csvout.'''
        csvout.writerow(self.__dict__)


# Traversal

def _next_or_None(it):
    '''Returns it.next(), or None if it is finished.'''
    try:
        return it.next()
    except StopIteration:
        return None


def step_up_walk(top, onerror=None):
    '''Like os.walk, but first yields from one step up. This is so that top can
    be treated once in the same way as everything else.

    top and onerror are the same as in os.walk().
    '''

    # Allow caller to remove top from dirs list and stop recursion
    dirs = [os.path.basename(top)]
    yield os.path.dirname(top), dirs, []
    if dirs:
        for e in os.walk(top, onerror=onerror):
            yield e


def path_lt(a, b):
    '''True if a will be examined before b in the walk of traverse().

    Files will not be traversed in simple lexicographic order, since all files
    in a directory will be examined before the contents of subdirectories.
    '''

    # Strip common prefix
    n = len(os.path.commonprefix([a, b]))
    a = a[n:]
    b = b[n:]

    # If either is directly in the lowest common dir, it will come first
    if '/' not in a and '/' in b:
        return True
    if '/' in a and '/' not in b:
        return False

    # Otherwise, compare the first differing path element, which, since we
    # stripped the common prefix, is the first path element in each.
    return a.split(os.sep, 1)[0] < b.split(os.sep, 1)[0]


def traverse(root, csvin, csvout, prune_cond=lambda x: False, gid=None):
    '''Traverse root, comparing its contents to records in csvin and recording
    its contents to csvout.

    prune_cond -- predicate of FileData; if true for a directory, it will be
                  pruned (i.e. its contents will not be inspected)

    gid        -- if not None, warn when a file's gid is different

    Returns a tuple (added, deleted, warnings).

    added    -- list of strings indicating files added since last run
    deleted  -- list of strings indicating files deleted since last run
    warnings -- {code: [mesgs]} for warnings encountered during traversal
    '''

    added = []
    deleted = []
    warnings = {}

    old_files = (FileData.fromfields(line) for line in csvin)
    # old is None once iteration is complete
    old = _next_or_None(old_files)

    # Function to pass to os.walk as onerror: add a warning
    def listdir_error(e):
        warnings.setdefault(W_OSERROR, []).append(
                "cannot listdir %s: %s" % (e.filename, e.strerror)
        )

    for curdir, dirs, files in step_up_walk(root, onerror=listdir_error):
        dirs.sort()  # sorts the dirs for the next level of recursion

        for name in sorted(files + dirs):
            try:
                f = FileData(os.path.join(curdir, name))
            except OSError as e:
                warnings.setdefault(W_OSERROR, []).append(
                        "cannot stat %s: %s" % (e.filename, e.strerror)
                )
                continue

            # Advance old until we find an old file corresponding to f, or
            # until we pass the place where it would be in the list
            while old is not None and path_lt(old.path, f.path):
                # old has disappeared since last time
                deleted.append(str(old))
                old = _next_or_None(old_files)

            # Add warnings
            for code, warnlist in f.get_warnings(gid=gid).iteritems():
                warnings.setdefault(code, []).extend(warnlist)

            # Prune if prune_cond says we should
            if f.isdir() and prune_cond(f):
                dirs.remove(os.path.basename(f.path))
                f.pruned = True

            # Add f to added if it didn't exist last time
            if old is None or old.path != f.path:
                added.append(str(f))
            else:
                # Check if f.pruned has changed since last time
                if f.pruned and not old.pruned:
                    warnings.setdefault(W_STARTEXCLUDE, []).append(str(f))
                elif not f.pruned and old.pruned:
                    warnings.setdefault(W_STOPEXCLUDE, []).append(str(f))
                # Advance old to avoid adding it to deleted on next iteration
                old = _next_or_None(old_files)

            f.tofile(csvout)

    # Add remaining old files to deleted
    if old is not None:
        deleted.append(str(old))
        deleted.extend(str(old) for old in old_files)

    return added, deleted, warnings


# Main and misc #

def get_notification(prevtime, added, deleted, warnings):
    '''Generates notification text.'''

    lines = []

    if prevtime is None:
        lines.append("First run for this directory.")
    else:
        ptime = datetime.fromtimestamp(prevtime).strftime('%Y-%m-%d %H:%M:%S')
        lines.append("Last run was %s." % ptime)

    lines.append("")

    # Added and deleted files
    for msg, lst in [
        ("Added since last run:", sorted(added)),
        ("Deleted since last run:", sorted(deleted)),
    ]:
        if lst:
            lines.append(msg)
            lines.extend(lst)
            lines.append("")

    # Warnings
    if warnings:
        # Track warning codes we encounter to print explanations later
        lines.append("Warnings:")

        for code, warnlist in sorted(warnings.items()):
            lines.append("%d %s:" % (code, W_MESSAGE[code]))
            lines.extend("  %s" % w for w in sorted(warnlist))
            lines.append("")

        # Explanations for warnings
        lines.append("Warning details:")
        for code in sorted(warnings.keys()):
            lines.append("%d %s\n%s\n" %
                         (code, W_MESSAGE[code], W_DETAIL[code]))

    # Add line used to invoke me
    lines.append("-- ")
    lines.append("Git Central Repository Monitor")
    lines.append(" ".join(sys.argv))

    return "\n".join(lines)


def parse_args(argv):
    '''Parses command-line options.'''
    parser = optparse.OptionParser(usage="usage: %prog -d dir [options]")
    parser.add_option("-d", "--directory", metavar="DIR",
                      help="target directory to examine (required)")
    parser.add_option("-g", "--group", metavar="GROUP",
                      action="append",
                      help="warn if a file doesn't belong one of GROUP(s). "
                           "Can be specified multiple times.",
                      default=[])
    parser.add_option("-e", "--exclude-regex", metavar="REGEX",
                      help="don't recurse into subdirectories whose paths "
                           "match REGEX (can be given multiple times)",
                      default=[], action='append')
    parser.add_option("--record-dir", metavar="DIR",
                      help="save records in DIR (default %default)",
                      default=DEFAULT_RECDIR)
    opts, args = parser.parse_args(argv[1:])

    if len(args) > 0:
        parser.error("too many arguments; run with --help")

    if not opts.directory:
        parser.error("target directory required (-d); run with --help")

    gids_or_groups = opts.group
    gid = None
    valid_gids = []
    for g in gids_or_groups:
        try:
            gid = int(g)
        except ValueError:  # group name
            try:
                gid = grp.getgrnam(g).gr_gid
            except KeyError:
                parser.error("group %s doesn't exist" % opts.group)
        else:  # g was a gid already check it a known group.
            try:
                grp.getgrgid(g)
            except KeyError:
                parser.error("group with gid %d doesn't exist" % opts.group)
        valid_gids.append(gid)
    # Overwrite opts.group as specified by user with a canonical gids.
    # Perhaps this is better done as an optparse action time?
    opts.group = valid_gids
    return opts, args


def main(argv):
    '''Sets up record files, starts traversal of the specified directory, and
    sends notification(s) for added/deleted files and warnings.'''

    opts, _ = parse_args(argv)

    # Setup

    # Open the old record file for reading
    directory = os.path.abspath(opts.directory)
    recfile = os.path.join(opts.record_dir, urllib.quote_plus(directory[1:]))
    try:
        prevtime = os.stat(recfile).st_ctime
        fin = open(recfile, 'rb')
    except (OSError, IOError) as e:
        if e.errno != errno.ENOENT:
            raise
        # Record file doesn't exist yet (e.g. first run)
        prevtime = None
        fin = open(os.devnull, 'rb')

    # Write the new record file to a temporary that will be moved in later.
    fdout, tmpfilename = tempfile.mkstemp(dir=opts.record_dir)
    fout = os.fdopen(fdout, 'wb')

    # Wrap fin and fout in CSV readers
    csvin = csv.DictReader(fin, fieldnames=FileData.FIELDS)
    csvout = csv.DictWriter(fout, fieldnames=FileData.FIELDS,
                            extrasaction='ignore')

    # Set up directory prune predicate for traverse.
    # Prune directory if it is a Git repo, or if its path matches any of the
    # regexes supplied on the command line.
    regexes = map(re.compile, opts.exclude_regex)

    def prune_cond(fdata):
        if fdata.type == 'git':
            return True
        for r in regexes:
            if r.search(fdata.path):
                return True
        return False

    # Traverse target directory
    added, deleted, warnings = traverse(
            directory, csvin, csvout,
            prune_cond=prune_cond, gid=opts.group
    )

    # Close files and move new record file into place
    fin.close()
    fout.close()
    os.rename(tmpfilename, recfile)

    # Print notifications
    if added or deleted or warnings:
        print(get_notification(prevtime, added, deleted, warnings))


if __name__ == '__main__':
    main(sys.argv)
